/*****************************************************************-*- c++ -*-*
  Be FileTypes and Attributes extention for Ruby , [be_stat] by tenmei
*****************************************************************************/
#define __BE_ATTR_CPP__
/*****************************************************************************
  INCLUDES
*****************************************************************************/
#include <Be.h>
#include <memory>
#include <cstdio>
#include <cstdlib>
#include <cstring>

#include "ruby.h"

#include "be_attr.h"

/*****************************************************************************
  DEFINES
*****************************************************************************/
using namespace std;
using namespace rb_be_attr;

/*****************************************************************************
  TYPES
*****************************************************************************/
namespace {

struct RbBeAttr {
    BNode     node;
    BNodeInfo node_info;
};

} // namespace

/*****************************************************************************
  PROTOTYPES
*****************************************************************************/

/*****************************************************************************
  EXTERNAL WORKS
*****************************************************************************/

/*****************************************************************************
  LOCAL WORKS
*****************************************************************************/

/*****************************************************************************
  PROCEDURES
*****************************************************************************/
namespace {

void free_BeAttr(RbBeAttr* p_stat)
{
    delete p_stat;
}

VALUE open_BeAttr(VALUE self,VALUE path)
{
    Check_Type(path,T_STRING);
    Check_SafeStr(path);
    
    auto_ptr<RbBeAttr> ap_stat(new RbBeAttr);
    
    if ((ap_stat->node.SetTo(STR2CSTR(path)) != B_OK) ||
        (ap_stat->node_info.SetTo(&(ap_stat->node)) != B_OK)) {
        rb_raise(rb_eRuntimeError,
                 "cannot found BNode for `%s\'!",STR2CSTR(path));
    }
    
    VALUE obj = Data_Wrap_Struct(self,0,
                                 reinterpret_cast<void(*)(...)>(free_BeAttr),
                                 ap_stat.release());
    
    return obj;
}

VALUE get_FileType(VALUE self)
{
    RbBeAttr* p_stat = static_cast<RbBeAttr*>(0);
    
    Data_Get_Struct(self,RbBeAttr,p_stat);
    
    char p_type[B_MIME_TYPE_LENGTH];
    
    if (p_stat->node_info.GetType(p_type) != B_OK) {
        rb_raise(rb_eRuntimeError,"cannot get FileType of Node!");
    }
    
    return rb_str_new2(p_type);
}

VALUE set_FileType(VALUE self,VALUE mime_str)
{
    Check_Type(mime_str,T_STRING);
    rb_secure(2);
    
    if (!BMimeType::IsValid(STR2CSTR(mime_str))) {
        rb_raise(rb_eRuntimeError,
                 "`%s\' is not valid MIME string!",STR2CSTR(mime_str));
    }
    
    RbBeAttr* p_stat = static_cast<RbBeAttr*>(0);
    
    Data_Get_Struct(self,RbBeAttr,p_stat);
    
    if (p_stat->node_info.SetType(STR2CSTR(mime_str)) != B_OK) {
        rb_raise(rb_eRuntimeError,"cannot set FileType of Node!");
    }
    
    return mime_str;
}

VALUE get_AttrNames(VALUE self)
{
    RbBeAttr* p_stat = static_cast<RbBeAttr*>(0);
    
    Data_Get_Struct(self,RbBeAttr,p_stat);
    
    char  name[B_ATTR_NAME_LENGTH];
    VALUE ret = rb_ary_new();
    
    p_stat->node.RewindAttrs();
    
    while (p_stat->node.GetNextAttrName(name) == B_OK) {
        rb_ary_push(ret,rb_str_new2(name));
    }
    
    return ret;
}

VALUE get_AttrInfo(VALUE self,VALUE name)
{
    Check_Type(name,T_STRING);
    
    RbBeAttr* p_stat = static_cast<RbBeAttr*>(0);
    
    Data_Get_Struct(self,RbBeAttr,p_stat);
    
    attr_info info;
    
    if (p_stat->node.GetAttrInfo(STR2CSTR(name),&info) != B_OK) {
        rb_raise(rb_eRuntimeError,
                 "cannot found Attribute , named `%s'!",STR2CSTR(name));
    }
    
    VALUE ret = rb_hash_new();
    
    rb_hash_aset(ret,rb_str_new2("type"),UINT2NUM(info.type));
    rb_hash_aset(ret,rb_str_new2("size"),INT2NUM(info.size));
    
    return ret;
}

template<typename T>
T read_Attr_Num(RbBeAttr* p_stat,const attr_info& info,VALUE name)
{
    T data;
    
    if (p_stat->node.ReadAttr(STR2CSTR(name),info.type,0,&data,sizeof(T)) <
        sizeof(T)) {
        rb_raise(rb_eRuntimeError,
                 "cannot read attribute data from  `%s'!",STR2CSTR(name));
    }
    
    return data;
}

char* read_Attr_Byte(RbBeAttr* p_stat,const attr_info& info,VALUE name,
                     char* p_data)
{
    if (p_stat->node.ReadAttr(STR2CSTR(name),info.type,0,p_data,info.size) <
        info.size) {
        rb_raise(rb_eRuntimeError,
                 "cannot read attribute data from  `%s'!",STR2CSTR(name));
    }
    
    return p_data;
}

VALUE read_Attr(VALUE self,VALUE name)
{
    Check_Type(name,T_STRING);
    
    RbBeAttr* p_stat = static_cast<RbBeAttr*>(0);
    
    Data_Get_Struct(self,RbBeAttr,p_stat);
    
    VALUE v_attr_info = get_AttrInfo(self,name);
    
    attr_info info;
    
    info.type = NUM2INT(rb_hash_aref(v_attr_info,rb_str_new2("type")));
    info.size = NUM2INT(rb_hash_aref(v_attr_info,rb_str_new2("size")));
    
    switch (info.type) {
    case B_BOOL_TYPE:
        if (read_Attr_Num<bool>(p_stat,info,name)) {
            return Qtrue;
        }
        return Qfalse;
        break;
    case B_CHAR_TYPE:
        return CHR2FIX(read_Attr_Num<char>(p_stat,info,name));
        break;
    case B_DOUBLE_TYPE:
        return rb_float_new(read_Attr_Num<double>(p_stat,info,name));
        break;
    case B_FLOAT_TYPE:
        return rb_float_new(
            static_cast<double>(read_Attr_Num<float>(p_stat,info,name)));
        break;
    case B_INT64_TYPE:
    {
        char  num_str[20];
        int64 num = read_Attr_Num<int64>(p_stat,info,name);
        
        if (num < 0LL) {
            num = -num;
            sprintf(num_str,"-%llx",num);
        } else {
            sprintf(num_str,"%llx",num);
        }
        
#if 0
        return rb_str2inum(num_str,16);
#else
        return rb_str2inum(rb_str_new2(num_str),16);
#endif
        break;
    }
    case B_INT32_TYPE:
    case B_SSIZE_T_TYPE:
        return INT2NUM(read_Attr_Num<int32>(p_stat,info,name));
        break;
    case B_INT16_TYPE:
        return INT2FIX(read_Attr_Num<int16>(p_stat,info,name));
        break;
    case B_INT8_TYPE:
#if 1
        return INT2FIX(read_Attr_Num<int8>(p_stat,info,name));
#else
        return CHR2FIX(read_Attr_Num<int8>(p_stat,info,name));
#endif
        break;
    case B_OFF_T_TYPE:
    case B_SIZE_T_TYPE:
    case B_UINT32_TYPE:
        return UINT2NUM(read_Attr_Num<uint32>(p_stat,info,name));
        break;
    case B_POINTER_TYPE:
        return UINT2NUM(
            reinterpret_cast<uint32>(read_Attr_Num<void*>(p_stat,info,name)));
        break;
    case B_UINT64_TYPE:
    {
        char  num_str[20];
        
        sprintf(num_str,"%llx",read_Attr_Num<int64>(p_stat,info,name));
        
#if 0
        return rb_str2inum(num_str,16);
#else
        return rb_str2inum(rb_str_new2(num_str),16);
#endif
        break;
    }
    case B_UINT16_TYPE:
        return UINT2NUM(read_Attr_Num<uint16>(p_stat,info,name));
        break;
    case B_UINT8_TYPE:
        return UINT2NUM(read_Attr_Num<uint8>(p_stat,info,name));
        break;
    case B_STRING_TYPE:
    {
        char* p_data = ALLOCA_N(char,info.size);
        return rb_str_new2(read_Attr_Byte(p_stat,info,name,p_data));
        break;
    }
    default:
    {
        char* p_data = ALLOCA_N(char,info.size);
        return rb_str_new(read_Attr_Byte(p_stat,info,name,p_data),info.size);
        break;
    }
    }
}

template<typename T>
void write_Attr_Num(RbBeAttr* p_stat,type_code code,const char* name,T data)
{
    if (p_stat->node.WriteAttr(name,code,0,&data,sizeof(T)) < sizeof(T)) {
        rb_raise(rb_eRuntimeError,
                 "cannot write attribute data to  `%s'!",STR2CSTR(name));
    }
}

void write_Attr_Byte(RbBeAttr* p_stat,type_code code,const char* name,
                     const void* p_buf,size_t length)
{
    if (p_stat->node.WriteAttr(name,code,0,p_buf,length) < length) {
        rb_raise(rb_eRuntimeError,
                 "cannot write attribute data to  `%s'!",STR2CSTR(name));
    }
}

VALUE write_Attr(VALUE self,VALUE name,VALUE type,VALUE data)
{
    Check_Type(name,T_STRING);
    rb_secure(2);
    
    RbBeAttr* p_stat = static_cast<RbBeAttr*>(0);
    
    Data_Get_Struct(self,RbBeAttr,p_stat);
    
    type_code code = NUM2ULONG(type);
    
    switch (code) {
    case B_BOOL_TYPE:
        if ((data == Qfalse) || (data == Qnil)) {
            write_Attr_Num<bool>(p_stat,code,STR2CSTR(name),false);
        } else {
            write_Attr_Num<bool>(p_stat,code,STR2CSTR(name),true);
        }
        break;
    case B_DOUBLE_TYPE:
        write_Attr_Num<double>(p_stat,code,STR2CSTR(name),NUM2DBL(data));
        break;
    case B_FLOAT_TYPE:
        write_Attr_Num<float>(p_stat,code,STR2CSTR(name),
                              static_cast<float>(NUM2DBL(data)));
        break;
    case B_INT64_TYPE:
    {
        VALUE num_str = rb_big2str(data,10);
        char* p_end;
        int64 num = strtoll(STR2CSTR(num_str),&p_end,10);
        
        write_Attr_Num<int64>(p_stat,code,STR2CSTR(name),num);
        break;
    }
    case B_INT32_TYPE:
    case B_SSIZE_T_TYPE:
        write_Attr_Num<int32>(p_stat,code,STR2CSTR(name),NUM2LONG(data));
        break;
    case B_INT16_TYPE:
        write_Attr_Num<int16>(p_stat,code,STR2CSTR(name),
                              static_cast<int16>(NUM2INT(data)));
        break;
    case B_INT8_TYPE:
        write_Attr_Num<int8>(p_stat,code,STR2CSTR(name),
                             static_cast<int8>(FIX2INT(data)));
        break;
    case B_OFF_T_TYPE:
    case B_SIZE_T_TYPE:
    case B_UINT32_TYPE:
        write_Attr_Num<uint32>(p_stat,code,STR2CSTR(name),NUM2ULONG(data));
        break;
    case B_POINTER_TYPE:
        write_Attr_Num<void*>(p_stat,code,STR2CSTR(name),
                              reinterpret_cast<void*>(NUM2ULONG(data)));
        break;
    case B_STRING_TYPE:
    {
        char*  p_str = STR2CSTR(data);
        size_t length = strlen(p_str) + 1;
        
        write_Attr_Byte(p_stat,code,STR2CSTR(name),
                        static_cast<void*>(p_str),length);
        break;
    }
    case B_UINT64_TYPE:
    {
        VALUE  num_str = rb_big2str(data,10);
        char*  p_end;
        uint64 num = strtoull(STR2CSTR(num_str),&p_end,10);
        
        write_Attr_Num<uint64>(p_stat,code,STR2CSTR(name),num);
        break;
    }
    case B_UINT16_TYPE:
        write_Attr_Num<uint16>(p_stat,code,STR2CSTR(name),
                               static_cast<uint16>(FIX2UINT(data)));
        break;
    case B_UINT8_TYPE:
        write_Attr_Num<uint8>(p_stat,code,STR2CSTR(name),
                              static_cast<uint8>(FIX2UINT(data)));
        break;
    default:
    {
        int   length;
        char* p_str = rb_str2cstr(data,&length);
        
        write_Attr_Byte(p_stat,code,STR2CSTR(name),
                        static_cast<void*>(p_str),
                        static_cast<size_t>(length));
        break;
    }
    }
    
    return name;
}

VALUE remove_Attr(VALUE self,VALUE name)
{
    Check_Type(name,T_STRING);
    rb_secure(2);
    
    RbBeAttr* p_stat = static_cast<RbBeAttr*>(0);
    
    Data_Get_Struct(self,RbBeAttr,p_stat);
    
    if (p_stat->node.RemoveAttr(STR2CSTR(name)) != B_OK) {
        rb_raise(rb_eRuntimeError,
                 "cannot remove attribute , named  `%s'!",STR2CSTR(name));
    }
    
    return name;
}

VALUE rename_Attr(VALUE self,VALUE name,VALUE new_name)
{
    Check_Type(name,T_STRING);
    Check_Type(new_name,T_STRING);
    rb_secure(2);
    
    RbBeAttr* p_stat = static_cast<RbBeAttr*>(0);
    
    Data_Get_Struct(self,RbBeAttr,p_stat);
    
#if 0 /* do not work ... why ? */
    if (p_stat->node.RenameAttr(STR2CSTR(name),STR2CSTR(new_name)) != B_OK) {
        rb_raise(rb_eRuntimeError,
                 "cannot rename attribute , named  `%s' to `%s'!",
                 STR2CSTR(name),STR2CSTR(new_name),-stat);
    }
#else /* use Read/WriteAttr instead of RenameAttr */
    if (strcmp(STR2CSTR(name),STR2CSTR(new_name)) == 0) return new_name;
    
    attr_info info;
    
    VALUE v_info = get_AttrInfo(self,name);
    
    info.type = NUM2INT(rb_hash_aref(v_info,rb_str_new2("type")));
    info.size = NUM2INT(rb_hash_aref(v_info,rb_str_new2("size")));
    
    void* p_buf = ALLOCA_N(char,info.size);
    
    if (p_stat->node.ReadAttr(STR2CSTR(name),info.type,0,p_buf,info.size) <
        info.size) {
        rb_raise(rb_eRuntimeError,
                 "cannot rename attribute , named  `%s' to `%s'!",
                 STR2CSTR(name),STR2CSTR(new_name));
    }
    if (p_stat->node.WriteAttr(STR2CSTR(new_name),
                               info.type,0,p_buf,info.size) < info.size) {
        rb_raise(rb_eRuntimeError,
                 "cannot rename attribute , named  `%s' to `%s'!",
                 STR2CSTR(name),STR2CSTR(new_name));
    }
    if (p_stat->node.RemoveAttr(STR2CSTR(name)) != B_OK) {
        rb_raise(rb_eRuntimeError,
                 "cannot rename attribute , named  `%s' to `%s'!",
                 STR2CSTR(name),STR2CSTR(new_name));
    }
#endif
    
    return new_name;
}

} // namespace

namespace rb_be_attr {

void init(VALUE BeStatVALUE)
{
    /*
     * Be Attribute class.
     * See BeBook StorageKit `BNode' and `BNodeInfo'.
     */
    VALUE BeAttrVALUE = rb_define_class_under(BeStatVALUE,"Attr",rb_cObject);
    
    rb_define_singleton_method(
        BeAttrVALUE,"new",reinterpret_cast<VALUE(*)(...)>(open_BeAttr),1);
    rb_define_singleton_method(
        BeAttrVALUE,"open",reinterpret_cast<VALUE(*)(...)>(open_BeAttr),1);
    
    rb_define_method(
        BeAttrVALUE,"GetType",
        reinterpret_cast<VALUE(*)(...)>(get_FileType),0);
    rb_define_method(
        BeAttrVALUE,"SetType",
        reinterpret_cast<VALUE(*)(...)>(set_FileType),1);
    rb_define_method(
        BeAttrVALUE,"GetNames",
        reinterpret_cast<VALUE(*)(...)>(get_AttrNames),0);
    rb_define_method(
        BeAttrVALUE,"GetInfo",
        reinterpret_cast<VALUE(*)(...)>(get_AttrInfo),1);
    rb_define_method(
        BeAttrVALUE,"Read",
        reinterpret_cast<VALUE(*)(...)>(read_Attr),1);
    rb_define_method(
        BeAttrVALUE,"Write",
        reinterpret_cast<VALUE(*)(...)>(write_Attr),3);
    rb_define_method(
        BeAttrVALUE,"Remove",
        reinterpret_cast<VALUE(*)(...)>(remove_Attr),1);
    rb_define_method(
        BeAttrVALUE,"Rename",
        reinterpret_cast<VALUE(*)(...)>(rename_Attr),2);
}

} // namespace rb_be_attr

/*****************************************************************************
  END OF FILE
*****************************************************************************/
